/*
 * @(#)RubikS.c
 *
 * Taken from code originally written by
 * Michael B. Martin <martinm@sps1.phys.vt.edu>
 * From cubist10.c-- for IBM PC.
 * Used by permission.
 * Taken from the algorithm in the Ideal Solution book.
 * Break ability taken from the X puzzle by Don Bennett, HP Labs
 *
 * Copyright 1994 - 2010  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

/* For break code */
/* Puzzle - (C) Copyright 1987, 1988 Don Bennett.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 */

/* Solver file for Rubik */

#define JMP
#ifdef JMP
#include <setjmp.h> /* longjmp ... interrupt */
#endif
#include "RubikP.h"

/*#define HANDLE_2 1*/

#define TOP_FACE 0
#define LEFT_FACE 1
#define FRONT_FACE 2
#define RIGHT_FACE 3
#define BOTTOM_FACE 4
#define BACK_FACE 5

#define I_TOP_FACE 5
#define I_LEFT_FACE 4
#define I_FRONT_FACE 1
#define I_RIGHT_FACE 2
#define I_BOTTOM_FACE 6
#define I_BACK_FACE 3

/* Mappings of Ideal notation to my General nxnxn cube notation */
#define rotateLeft(w)	movePuzzlePiece(w,FRONT_FACE,0,LEFT,TRUE)
#define rotateRight(w)	movePuzzlePiece(w,FRONT_FACE,0,RIGHT,TRUE)
#define rotateUp(w)	movePuzzlePiece(w,FRONT_FACE,0,TOP,TRUE)
#define rotateDown(w)	movePuzzlePiece(w,FRONT_FACE,0,BOTTOM,TRUE)
#define rotateCw(w)	movePuzzlePiece(w,TOP_FACE,0,RIGHT,TRUE)
#define rotateCcw(w)	movePuzzlePiece(w,TOP_FACE,0,LEFT,TRUE)

#define REV (w->rubik.sizex-1)
#define rotateTopLeft(w)	movePuzzlePiece(w,FRONT_FACE,0,LEFT,FALSE)
#define rotateCenterLeft(w)	movePuzzlePiece(w,FRONT_FACE,w->rubik.sizex,LEFT,FALSE)
#define rotateBottomLeft(w)	movePuzzlePiece(w,FRONT_FACE,w->rubik.sizex*REV,LEFT,FALSE)
#define rotateTopRight(w)	movePuzzlePiece(w,FRONT_FACE,0,RIGHT,FALSE)
#define rotateCenterRight(w)	movePuzzlePiece(w,FRONT_FACE,w->rubik.sizex,RIGHT,FALSE)
#define rotateBottomRight(w)	movePuzzlePiece(w,FRONT_FACE,w->rubik.sizex*REV,RIGHT,FALSE)
#define rotateLeftUp(w)		movePuzzlePiece(w,FRONT_FACE,0,TOP,FALSE)
#define rotateCenterUp(w)	movePuzzlePiece(w,FRONT_FACE,1,TOP,FALSE)
#define rotateRightUp(w)	movePuzzlePiece(w,FRONT_FACE,REV,TOP,FALSE)
#define rotateLeftDown(w)	movePuzzlePiece(w,FRONT_FACE,0,BOTTOM,FALSE)
#define rotateCenterDown(w)	movePuzzlePiece(w,FRONT_FACE,1,BOTTOM,FALSE)
#define rotateRightDown(w)	movePuzzlePiece(w,FRONT_FACE,REV,BOTTOM,FALSE)
#define rotateFrontCw(w)	movePuzzlePiece(w,TOP_FACE,w->rubik.sizex*REV,RIGHT,FALSE)
#define rotateFrontCcw(w)	movePuzzlePiece(w,TOP_FACE,w->rubik.sizex*REV,LEFT,FALSE)

/* Still needs some work */
#ifdef REVENGE
/* Revenge start from 0 instead of 1 */

/* Revenge Table Moves */
#define rotateTableLeft(w,table)	movePuzzlePiece(w,FRONT_FACE,(table)*w->rubik.sizex,LEFT,FALSE)
#define rotateTableRight(w,table)	movePuzzlePiece(w,FRONT_FACE,(table)*w->rubik.sizex,RIGHT,FALSE)

/* Revenge Book Moves */
#define rotateBookDown(w,book)	movePuzzlePiece(w,FRONT_FACE,book,BOTTOM,FALSE)
#define rotateBookUp(w,book)	movePuzzlePiece(w,FRONT_FACE,book,TOP,FALSE)

/* Revenge Window Moves */
#define rotateWindowCcw(w,window)	movePuzzlePiece(w,TOP_FACE,window,BOTTOM,FALSE)
#define rotateWindowCw(w,window)	movePuzzlePiece(w,TOP_FACE,window,TOP,FALSE)


typedef struct _RubikTBW {
	int table, book, window;
} RubikTBW;

typedef struct _RubikFP {
	int face, x, y;
} RubikFP;

RubikTBW rubikTbw;
RubikFP rubikFp;

static void printTBW(RubikTBW tbw) {
	(void) printf("  table=%d book=%d window=%d\n",
		tbw.table, tbw.book, tbw.window);
}

static void printFP(RubikFP fp) {
	(void) printf("  fp.face=%d position=(%d, %d)\n",
		fp.face, fp.x, fp.y);
}

static RubikTBW mapToTBW(RubikWidget w, int face, int x, int y)
{
	switch (face) {
	case 0:
		rubikTbw.table = 0;
		rubikTbw.book = x;
		rubikTbw.window = REV - y;
		return rubikTbw;
	case 1:
		rubikTbw.table = y;
		rubikTbw.book = 0;
		rubikTbw.window = REV - x;
		return rubikTbw;
	case 2:
		rubikTbw.table = y;
		rubikTbw.book = x;
		rubikTbw.window = 0;
		return rubikTbw;
	case 3:
		rubikTbw.table = y;
		rubikTbw.book = REV;
		rubikTbw.window = x;
		return rubikTbw;
	case 4:
		rubikTbw.table = REV;
		rubikTbw.book = x;
		rubikTbw.window = y;
		return rubikTbw;
	case 5:
		rubikTbw.table = y;
		rubikTbw.book = REV - x;
		rubikTbw.window = REV;
		return rubikTbw;
	default:
		return rubikTbw;
	}
}

static RubikFP mapFromTBW(RubikWidget w, int table, int book, int window)
{
	if (table == 0) {
		rubikFp.face = 0;
		rubikFp.x = book;
		rubikFp.y = REV - window;
		return rubikFp;
	} else if (table == REV) {
		rubikFp.face = 4;
		rubikFp.x = book;
		rubikFp.y = window;
		return rubikFp;
	}
	if (book == 0) {
		rubikFp.face = 1;
		rubikFp.y = table;
		rubikFp.x = REV - window;
		return rubikFp;
	} else if (book == REV) {
		rubikFp.face = 3;
		rubikFp.y = table;
		rubikFp.x = window;
		return rubikFp;
	}
	if (window == 0) {
		rubikFp.face = 2;
		rubikFp.y = table;
		rubikFp.x = book;
		return rubikFp;
	} else if (window == REV) {
		rubikFp.face = 5;
		rubikFp.y = table;
		rubikFp.x = REV - book;
		return rubikFp;
	}
	return rubikFp;
}
#endif

#ifdef DEBUG
static int mapGeneralToIdeal[MAX_FACES] =
{5, 4, 1, 2, 6, 3};

#endif
static int mapIdealToGeneral[MAX_FACES] =
{2, 3, 5, 1, 0, 4};		/* Remember to subtract 1 */

static Boolean solvingFlag = False;

#ifdef JMP
static Boolean abortSolvingFlag = False;
static jmp_buf solve_env;

static void
abortSolving(void)
{
	if (solvingFlag)
		abortSolvingFlag = True;
}

#ifdef WINVER
static Boolean
processMessage(UINT msg)
{
	switch (msg) {
	case WM_KEYDOWN:
	case WM_CLOSE:
	case WM_LBUTTONDOWN:
	case WM_RBUTTONDOWN:
		abortSolving();
		return True;
	default:
		return False;
	}
}
#else
static void
processButton(void /*XButtonEvent *event*/)
{
	abortSolving();
}

static void
processVisibility(XVisibilityEvent *event)
{
	if (event->state != VisibilityUnobscured)
		abortSolving();
}

static void
getNextEvent(RubikWidget w, XEvent *event)
{
	if (!XCheckMaskEvent(XtDisplay(w), VisibilityChangeMask, event))
		(void) XNextEvent(XtDisplay(w), event);
}

static void
processEvent(XEvent *event)
{
	switch(event->type) {
	case KeyPress:
	case ButtonPress:
		processButton(/*&event->xbutton*/);
		break;
	case VisibilityNotify:
		processVisibility(&event->xvisibility);
		break;
	default:
		break;
	}
}

static void
processEvents(RubikWidget w)
{
	XEvent event;

	while (XPending(XtDisplay(w))) {
		getNextEvent(w, &event);
		processEvent(&event);
	}
}
#endif
#endif

static void
movePuzzlePiece(RubikWidget w,
	const int face, const int position, const int direction,
	const int control)
{
#ifdef JMP
#ifdef WINVER
	MSG msg;

	if (PeekMessage(&msg, NULL, 0, 0, 0)) {
		if (!processMessage(msg.message)) {
			if (GetMessage(&msg, NULL, 0, 0))
				DispatchMessage(&msg);
		}
	}
#else
	processEvents(w);
#endif
	if (solvingFlag && abortSolvingFlag)
		longjmp(solve_env, 1);
#endif
	movePuzzleDelay(w, face, position, direction, control);
}

static int cornerFace(RubikWidget w, int quad) {
#ifdef HANDLE_2
	if (w->rubik.sizex == 2) {
		switch (quad) {
		case 0:
			return 0;
		case 1:
			return 2;
		case 2:
			return 6;
		case 3:
			return 8;
		default:
			return 9;
		}
	}
#endif
	/* top left to bottom right */
	switch (quad) {
	case 0:
		return 0;
	case 1:
		return REV;
	case 2:
		return w->rubik.sizex * REV;
	case 3:
		return w->rubik.sizex * w->rubik.sizex - 1;
	default:
		return w->rubik.sizex * w->rubik.sizex;
	}
}

/* For size 3
 *  0        1
 * 1 2  =>  3 5
 *  3        7
 */
static int sideFace(RubikWidget w, int quad) {
	/* top left to bottom right */
	/* if even return lowest */
	switch (quad) {
	case 0:
		return REV / 2;
	case 1:
		return (REV / 2) * w->rubik.sizex;
	case 2:
		return (REV / 2) * w->rubik.sizex + REV;
	case 3:
		return w->rubik.sizex * w->rubik.sizex
			- w->rubik.sizex / 2 - 1;
	default:
		return w->rubik.sizex * w->rubik.sizex;
	}
}

static int centerFace(RubikWidget w) {
	/* if even return lowest */
	return (REV / 2) * w->rubik.sizex + REV / 2;
}

static int
colorSide(RubikWidget w, int m, int c)
{
	int d, i, j;

	d = mapIdealToGeneral[m - 1];
#ifdef HANDLE_2
	i = c % 3;
	j = c / 3;
	if (i == 2)
		i = w->rubik.sizex - 1;
	if (j == 2)
		j = w->rubik.sizex - 1;
#else
	i = c % w->rubik.sizex;
	j = c / w->rubik.sizex;
#endif
	return w->rubik.cubeLoc[d][j * w->rubik.sizex + i].face;
}

static int
mapFace(RubikWidget w, int m)
{
	int d;

	d = mapIdealToGeneral[m - 1];
	return ((w->rubik.cubeLoc[d][4].rotation -
		w->rubik.cubeLoc[d][0].rotation + MAX_ORIENT) % MAX_ORIENT);
}

/* This procedure finds the location of a specified corner.  An */
/* improperly-specified color combination will result in a return */
/* value of 9. */
static int
findCorner(RubikWidget w, int color1, int color2, int color3)
{
	int corner = 0, temp = 1;

	do {
		if (colorSide(w, 5, cornerFace(w, 2)) == color1) {
			if (colorSide(w, 1, cornerFace(w, 0)) == color2 &&
					colorSide(w, 4, cornerFace(w, 1)) == color3)
				corner = temp;
		} else if (colorSide(w, 5, cornerFace(w, 2)) == color2) {
			if (colorSide(w, 1, cornerFace(w, 0)) == color3 &&
					colorSide(w, 4, cornerFace(w, 1)) == color1)
				corner = temp;
		} else if (colorSide(w, 5, cornerFace(w, 2)) == color3 &&
				colorSide(w, 1, cornerFace(w, 0)) == color1 &&
				colorSide(w, 4, cornerFace(w, 1)) == color2) {
			corner = temp;
		}
		if (corner == 0) {
			if (temp < 4)
				rotateLeft(w);
			else if (temp == 4) {
				rotateCw(w);
				rotateCw(w);
			} else if (temp < 8)
				rotateRight(w);
			else if (temp == 8)
				corner = 9;
			temp++;
		}
	} while (corner == 0);

	/* put the cube back to the way it was */
	if (corner == 2)
		rotateRight(w);
	else if (corner == 3) {
		rotateRight(w);
		rotateRight(w);
	} else if (corner == 4)
		rotateLeft(w);
	else if (corner == 5) {
		rotateCw(w);
		rotateCw(w);
		rotateLeft(w);
	} else if (corner == 6) {
		rotateCw(w);
		rotateCw(w);
	} else if (corner == 7) {
		rotateCw(w);
		rotateCw(w);
		rotateRight(w);
	} else if (corner == 8) {
		rotateCw(w);
		rotateCw(w);
		rotateRight(w);
		rotateRight(w);
	}
	return (corner);
}

/* This procedure finds the location of a specified edge.  An */
/* improperly-specified color combination will result in a return */
/* value of 13. */
static int
findEdge(RubikWidget w, int color1, int color2)
{
	int edge = 0, temp = 1;

	do {
		/* Top (5) down & Front (1) top edge */
		if (colorSide(w, 5, sideFace(w, 3)) == color1) {
			if (colorSide(w, 1, sideFace(w, 0)) == color2)
				edge = temp;
		} else if (colorSide(w, 5, sideFace(w, 3)) == color2 &&
				colorSide(w, 1, sideFace(w, 0)) == color1) {
			edge = temp;
		}
		if (edge == 0) {
			if (temp < 4)
				rotateLeft(w);
			else if (temp == 4) {
				rotateLeft(w);
				rotateCw(w);
			} else if (temp < 8)
				rotateUp(w);
			else if (temp == 8) {
				rotateUp(w);
				rotateCw(w);
			} else if (temp < 12)
				rotateRight(w);
			else if (temp == 12)
				edge = 13;	/* end condition, just in case */
			temp++;
		}
	} while (edge == 0);

	/* put the cube back to the way it was */
	if (edge == 2) {
		rotateRight(w);
	} else if (edge == 3) {
		rotateRight(w);
		rotateRight(w);
	} else if (edge == 4) {
		rotateLeft(w);
	} else if (edge == 5) {
		rotateCcw(w);
	} else if (edge == 6) {
		rotateCcw(w);
		rotateRight(w);
	} else if (edge == 7) {
		rotateCcw(w);
		rotateRight(w);
		rotateRight(w);
	} else if (edge == 8) {
		rotateCcw(w);
		rotateLeft(w);
	} else if (edge == 9) {
		rotateCcw(w);
		rotateCcw(w);
	} else if (edge == 10) {
		rotateCcw(w);
		rotateCcw(w);
		rotateRight(w);
	} else if (edge == 11) {
		rotateUp(w);
		rotateUp(w);
	} else if (edge == 12) {
		rotateUp(w);
		rotateUp(w);
		rotateRight(w);
	}
	return (edge);
}

/* This procedure places the specified edge piece in edge position */
/* #1 (front top). */
static void
placeEdge(RubikWidget w, int edge, int top_color)
{
	/* first put the edge piece in position #8 (center row, left rear) */
	if (edge == 1) {
		if (colorSide(w, 5, sideFace(w, 3)) == top_color)
			return;	/* already in place */
		else {
			rotateFrontCcw(w);
			rotateCenterLeft(w);
			rotateFrontCw(w);
		}
	} else if (edge == 2) {
		rotateTopLeft(w);
		rotateFrontCcw(w);
		rotateCenterLeft(w);
		rotateFrontCw(w);
		rotateTopRight(w);
	} else if (edge == 3) {
		rotateTopLeft(w);
		rotateTopLeft(w);
		rotateFrontCcw(w);
		rotateCenterLeft(w);
		rotateFrontCw(w);
		rotateTopRight(w);
		rotateTopRight(w);
	} else if (edge == 4) {
		rotateTopRight(w);
		rotateFrontCcw(w);
		rotateCenterLeft(w);
		rotateFrontCw(w);
		rotateTopLeft(w);
	} else if (edge == 5) {
		rotateCenterLeft(w);
	} else if (edge == 6) {
		rotateCenterLeft(w);
		rotateCenterLeft(w);
	} else if (edge == 7) {
		rotateCenterRight(w);
	} else if (edge == 9) {
		rotateFrontCw(w);
		rotateCenterLeft(w);
		rotateFrontCcw(w);
	} else if (edge == 10) {
		rotateBottomLeft(w);
		rotateFrontCw(w);
		rotateCenterLeft(w);
		rotateFrontCcw(w);
		rotateBottomRight(w);
	} else if (edge == 11) {
		rotateBottomLeft(w);
		rotateBottomLeft(w);
		rotateFrontCw(w);
		rotateCenterLeft(w);
		rotateFrontCcw(w);
		rotateBottomRight(w);
		rotateBottomRight(w);
	} else if (edge == 12) {
		rotateBottomRight(w);
		rotateFrontCw(w);
		rotateCenterLeft(w);
		rotateFrontCcw(w);
		rotateBottomLeft(w);
	}
	/* put the piece in place */
	if (colorSide(w, 4, sideFace(w, 1)) == top_color) {
		rotateFrontCw(w);
		rotateCenterRight(w);
		rotateCenterRight(w);
		rotateFrontCcw(w);
	} else {
		rotateFrontCcw(w);
		rotateCenterRight(w);
		rotateFrontCw(w);
	}
}

/* This procedure solves the top (TOP_FACE) side, except for one edge. */
static void
solveTop(RubikWidget w)
{
	int corner, edge;

	/* put the blue face on the top */
	if (colorSide(w, 1, centerFace(w)) == TOP_FACE)
		rotateUp(w);
	else if (colorSide(w, 2, centerFace(w)) == TOP_FACE)
		rotateCcw(w);
	else if (colorSide(w, 3, centerFace(w)) == TOP_FACE)
		rotateDown(w);
	else if (colorSide(w, 4, centerFace(w)) == TOP_FACE)
		rotateCw(w);
	else if (colorSide(w, 6, centerFace(w)) == TOP_FACE) {
		rotateUp(w);
		rotateUp(w);
	}
	/* first find the blue-red-white corner and place it */
	corner = findCorner(w, TOP_FACE, FRONT_FACE, LEFT_FACE);
	if (corner == 1) {
		if (colorSide(w, 5, cornerFace(w, 2)) == FRONT_FACE) {
			rotateFrontCw(w);
			rotateTopLeft(w);
		} else if (colorSide(w, 5, cornerFace(w, 2)) == LEFT_FACE) {
			rotateLeftUp(w);
			rotateTopRight(w);
		}
	} else if (corner == 2) {
		if (colorSide(w, 5, cornerFace(w, 3)) == TOP_FACE)
			rotateTopLeft(w);
		else if (colorSide(w, 5, cornerFace(w, 3)) == FRONT_FACE) {
			rotateRightUp(w);
			rotateTopLeft(w);
			rotateTopLeft(w);
		} else if (colorSide(w, 5, cornerFace(w, 3)) == LEFT_FACE)
			rotateFrontCcw(w);
	} else if (corner == 3) {
		if (colorSide(w, 5, cornerFace(w, 1)) == TOP_FACE) {
			rotateTopLeft(w);
			rotateTopLeft(w);
		} else if (colorSide(w, 5, cornerFace(w, 1)) == FRONT_FACE) {
			rotateTopRight(w);
			rotateLeftDown(w);
		} else if (colorSide(w, 5, cornerFace(w, 1)) == LEFT_FACE) {
			rotateRightDown(w);
			rotateTopLeft(w);
		}
	} else if (corner == 4) {
		if (colorSide(w, 5, cornerFace(w, 0)) == TOP_FACE)
			rotateTopRight(w);
		else if (colorSide(w, 5, cornerFace(w, 0)) == FRONT_FACE)
			rotateLeftDown(w);
		else if (colorSide(w, 5, cornerFace(w, 0)) == LEFT_FACE) {
			rotateTopRight(w);
			rotateLeftUp(w);
			rotateTopRight(w);
		}
	} else if (corner == 5) {
		if (colorSide(w, 6, cornerFace(w, 0)) == TOP_FACE) {
			rotateBottomLeft(w);
			rotateLeftUp(w);
			rotateLeftUp(w);
		} else if (colorSide(w, 6, cornerFace(w, 0)) == FRONT_FACE)
			rotateLeftUp(w);
		else if (colorSide(w, 6, cornerFace(w, 0)) == LEFT_FACE)
			rotateFrontCw(w);
	} else if (corner == 6) {
		if (colorSide(w, 6, cornerFace(w, 1)) == TOP_FACE) {
			rotateFrontCw(w);
			rotateFrontCw(w);
		} else if (colorSide(w, 6, cornerFace(w, 1)) == FRONT_FACE) {
			rotateBottomLeft(w);
			rotateLeftUp(w);
		} else if (colorSide(w, 6, cornerFace(w, 1)) == LEFT_FACE) {
			rotateBottomLeft(w);
			rotateFrontCw(w);
		}
	} else if (corner == 7) {
		if (colorSide(w, 6, cornerFace(w, 3)) == TOP_FACE) {
			rotateBottomLeft(w);
			rotateFrontCw(w);
			rotateFrontCw(w);
		} else if (colorSide(w, 6, cornerFace(w, 3)) == FRONT_FACE) {
			rotateBottomLeft(w);
			rotateBottomLeft(w);
			rotateLeftUp(w);
		} else if (colorSide(w, 6, cornerFace(w, 3)) == LEFT_FACE) {
			rotateBottomLeft(w);
			rotateBottomLeft(w);
			rotateFrontCw(w);
		}
	} else if (corner == 8) {
		if (colorSide(w, 6, cornerFace(w, 2)) == TOP_FACE) {
			rotateLeftUp(w);
			rotateLeftUp(w);
		} else if (colorSide(w, 6, cornerFace(w, 2)) == FRONT_FACE) {
			rotateBottomRight(w);
			rotateLeftUp(w);
		} else if (colorSide(w, 6, cornerFace(w, 2)) == LEFT_FACE) {
			rotateBottomRight(w);
			rotateFrontCw(w);
		}
	}
	/* now find the blue-yellow-red corner and place it */
	rotateLeft(w);
	corner = findCorner(w, TOP_FACE, RIGHT_FACE, FRONT_FACE);
	if (corner == 1) {
		if (colorSide(w, 5, cornerFace(w, 2)) == RIGHT_FACE) {
			rotateFrontCcw(w);
			rotateBottomRight(w);
			rotateFrontCcw(w);
			rotateFrontCcw(w);
		} else if (colorSide(w, 5, cornerFace(w, 2)) == FRONT_FACE) {
			rotateFrontCw(w);
			rotateFrontCw(w);
			rotateBottomLeft(w);
			rotateFrontCw(w);
		}
	} else if (corner == 2) {
		if (colorSide(w, 5, cornerFace(w, 3)) == TOP_FACE) {
			rotateRightDown(w);
			rotateBottomLeft(w);
			rotateFrontCw(w);
		} else if (colorSide(w, 5, cornerFace(w, 3)) == RIGHT_FACE) {
			rotateRightDown(w);
			rotateFrontCw(w);
			rotateFrontCw(w);
		} else if (colorSide(w, 5, cornerFace(w, 3)) == FRONT_FACE)
			rotateFrontCcw(w);
	} else if (corner == 3) {
		if (colorSide(w, 5, cornerFace(w, 1)) == TOP_FACE) {
			rotateRightDown(w);
			rotateRightDown(w);
			rotateFrontCw(w);
			rotateFrontCw(w);
		} else if (colorSide(w, 5, cornerFace(w, 1)) == RIGHT_FACE) {
			rotateRightDown(w);
			rotateFrontCcw(w);
		} else if (colorSide(w, 5, cornerFace(w, 1)) == FRONT_FACE) {
			rotateRightDown(w);
			rotateRightDown(w);
			rotateBottomLeft(w);
			rotateFrontCw(w);
		}
	} else if (corner == 5) {
		if (colorSide(w, 6, cornerFace(w, 0)) == TOP_FACE) {
			rotateBottomRight(w);
			rotateFrontCw(w);
			rotateFrontCw(w);
		} else if (colorSide(w, 6, cornerFace(w, 0)) == RIGHT_FACE) {
			rotateBottomRight(w);
			rotateRightUp(w);
			rotateFrontCcw(w);
		} else if (colorSide(w, 6, cornerFace(w, 0)) == FRONT_FACE)
			rotateFrontCw(w);
	} else if (corner == 6) {
		if (colorSide(w, 6, cornerFace(w, 1)) == TOP_FACE) {
			rotateFrontCw(w);
			rotateFrontCw(w);
		} else if (colorSide(w, 6, cornerFace(w, 1)) == RIGHT_FACE) {
			rotateRightUp(w);
			rotateFrontCcw(w);
		} else if (colorSide(w, 6, cornerFace(w, 1)) == FRONT_FACE) {
			rotateBottomLeft(w);
			rotateFrontCw(w);
		}
	} else if (corner == 7) {
		if (colorSide(w, 6, cornerFace(w, 3)) == TOP_FACE) {
			rotateBottomLeft(w);
			rotateFrontCw(w);
			rotateFrontCw(w);
		} else if (colorSide(w, 6, cornerFace(w, 3)) == RIGHT_FACE) {
			rotateBottomLeft(w);
			rotateRightUp(w);
			rotateFrontCcw(w);
		} else if (colorSide(w, 6, cornerFace(w, 3)) == FRONT_FACE) {
			rotateBottomLeft(w);
			rotateBottomLeft(w);
			rotateFrontCw(w);
		}
	} else if (corner == 8) {
		if (colorSide(w, 6, cornerFace(w, 2)) == TOP_FACE) {
			rotateBottomRight(w);
			rotateBottomRight(w);
			rotateFrontCw(w);
			rotateFrontCw(w);
		} else if (colorSide(w, 6, cornerFace(w, 2)) == RIGHT_FACE) {
			rotateBottomRight(w);
			rotateBottomRight(w);
			rotateRightUp(w);
			rotateFrontCcw(w);
		} else if (colorSide(w, 6, cornerFace(w, 2)) == FRONT_FACE) {
			rotateBottomRight(w);
			rotateFrontCw(w);
		}
	}
	/* now find the blue-orange-yellow corner and place it */
	rotateLeft(w);
	corner = findCorner(w, TOP_FACE, BACK_FACE, RIGHT_FACE);
	if (corner == 1) {
		if (colorSide(w, 5, cornerFace(w, 2)) == BACK_FACE) {
			rotateFrontCcw(w);
			rotateBottomRight(w);
			rotateFrontCcw(w);
			rotateFrontCcw(w);
		} else if (colorSide(w, 5, cornerFace(w, 2)) == RIGHT_FACE) {
			rotateFrontCw(w);
			rotateFrontCw(w);
			rotateBottomLeft(w);
			rotateFrontCw(w);
		}
	} else if (corner == 2) {
		if (colorSide(w, 5, cornerFace(w, 3)) == TOP_FACE) {
			rotateRightDown(w);
			rotateBottomLeft(w);
			rotateRightUp(w);
			rotateFrontCw(w);
		} else if (colorSide(w, 5, cornerFace(w, 3)) == BACK_FACE) {
			rotateFrontCw(w);
			rotateBottomLeft(w);
			rotateFrontCw(w);
		} else if (colorSide(w, 5, cornerFace(w, 3)) == RIGHT_FACE)
			rotateFrontCcw(w);
	} else if (corner == 5) {
		if (colorSide(w, 6, cornerFace(w, 0)) == TOP_FACE) {
			rotateBottomRight(w);
			rotateFrontCw(w);
			rotateFrontCw(w);
		} else if (colorSide(w, 6, cornerFace(w, 0)) == BACK_FACE) {
			rotateRightDown(w);
			rotateBottomRight(w);
			rotateRightUp(w);
			rotateFrontCcw(w);
		} else if (colorSide(w, 6, cornerFace(w, 0)) == RIGHT_FACE)
			rotateFrontCw(w);
	} else if (corner == 6) {
		if (colorSide(w, 6, cornerFace(w, 1)) == TOP_FACE) {
			rotateFrontCw(w);
			rotateFrontCw(w);
		} else if (colorSide(w, 6, cornerFace(w, 1)) == BACK_FACE) {
			rotateBottomLeft(w);
			rotateRightDown(w);
			rotateBottomRight(w);
			rotateRightUp(w);
			rotateFrontCcw(w);
		} else if (colorSide(w, 6, cornerFace(w, 1)) == RIGHT_FACE) {
			rotateBottomLeft(w);
			rotateFrontCw(w);
		}
	} else if (corner == 7) {
		if (colorSide(w, 6, cornerFace(w, 3)) == TOP_FACE) {
			rotateBottomLeft(w);
			rotateFrontCw(w);
			rotateFrontCw(w);
		} else if (colorSide(w, 6, cornerFace(w, 3)) == BACK_FACE) {
			rotateBottomLeft(w);
			rotateBottomLeft(w);
			rotateRightDown(w);
			rotateBottomRight(w);
			rotateRightUp(w);
			rotateFrontCcw(w);
		} else if (colorSide(w, 6, cornerFace(w, 3)) == RIGHT_FACE) {
			rotateBottomLeft(w);
			rotateBottomLeft(w);
			rotateFrontCw(w);
		}
	} else if (corner == 8) {
		if (colorSide(w, 6, cornerFace(w, 2)) == TOP_FACE) {
			rotateBottomRight(w);
			rotateBottomRight(w);
			rotateFrontCw(w);
			rotateFrontCw(w);
		} else if (colorSide(w, 6, cornerFace(w, 2)) == BACK_FACE) {
			rotateRightDown(w);
			rotateBottomRight(w);
			rotateBottomRight(w);
			rotateRightUp(w);
			rotateFrontCcw(w);
		} else if (colorSide(w, 6, cornerFace(w, 2)) == RIGHT_FACE) {
			rotateBottomRight(w);
			rotateFrontCw(w);
		}
	}
	/* and now find the blue-white-orange corner and place it */
	rotateLeft(w);
	corner = findCorner(w, TOP_FACE, LEFT_FACE, BACK_FACE);
	if (corner == 1) {
		if (colorSide(w, 5, cornerFace(w, 2)) == LEFT_FACE) {
			rotateLeftDown(w);
			rotateBottomRight(w);
			rotateLeftUp(w);
			rotateBottomLeft(w);
			rotateBottomLeft(w);
			rotateFrontCcw(w);
			rotateBottomRight(w);
			rotateFrontCw(w);
		} else if (colorSide(w, 5, cornerFace(w, 2)) == BACK_FACE) {
			rotateFrontCcw(w);
			rotateBottomLeft(w);
			rotateFrontCw(w);
			rotateBottomRight(w);
			rotateBottomRight(w);
			rotateLeftDown(w);
			rotateBottomLeft(w);
			rotateLeftUp(w);
		}
	} else if (corner == 5) {
		if (colorSide(w, 6, cornerFace(w, 0)) == TOP_FACE) {
			rotateBottomRight(w);
			rotateLeftDown(w);
			rotateBottomLeft(w);
			rotateBottomLeft(w);
			rotateLeftUp(w);
			rotateBottomRight(w);
			rotateLeftDown(w);
			rotateBottomLeft(w);
			rotateLeftUp(w);
		} else if (colorSide(w, 6, cornerFace(w, 0)) == LEFT_FACE) {
			rotateBottomRight(w);
			rotateLeftDown(w);
			rotateBottomLeft(w);
			rotateLeftUp(w);
		} else if (colorSide(w, 6, cornerFace(w, 0)) == BACK_FACE) {
			rotateBottomLeft(w);
			rotateFrontCcw(w);
			rotateBottomRight(w);
			rotateFrontCw(w);
		}
	} else if (corner == 6) {
		if (colorSide(w, 6, cornerFace(w, 1)) == TOP_FACE) {
			rotateBottomRight(w);
			rotateFrontCcw(w);
			rotateBottomLeft(w);
			rotateFrontCw(w);
			rotateBottomLeft(w);
			rotateFrontCcw(w);
			rotateBottomRight(w);
			rotateFrontCw(w);
		} else if (colorSide(w, 6, cornerFace(w, 1)) == LEFT_FACE) {
			rotateLeftDown(w);
			rotateBottomLeft(w);
			rotateLeftUp(w);
		} else if (colorSide(w, 6, cornerFace(w, 1)) == BACK_FACE) {
			rotateBottomLeft(w);
			rotateBottomLeft(w);
			rotateFrontCcw(w);
			rotateBottomRight(w);
			rotateFrontCw(w);
		}
	} else if (corner == 7) {
		if (colorSide(w, 6, cornerFace(w, 3)) == TOP_FACE) {
			rotateLeftDown(w);
			rotateBottomRight(w);
			rotateLeftUp(w);
			rotateBottomRight(w);
			rotateLeftDown(w);
			rotateBottomLeft(w);
			rotateLeftUp(w);
		} else if (colorSide(w, 6, cornerFace(w, 3)) == LEFT_FACE) {
			rotateLeftDown(w);
			rotateBottomLeft(w);
			rotateBottomLeft(w);
			rotateLeftUp(w);
		} else if (colorSide(w, 6, cornerFace(w, 3)) == BACK_FACE) {
			rotateFrontCcw(w);
			rotateBottomLeft(w);
			rotateBottomLeft(w);
			rotateFrontCw(w);
		}
	} else if (corner == 8) {
		if (colorSide(w, 6, cornerFace(w, 2)) == TOP_FACE) {
			rotateFrontCcw(w);
			rotateBottomRight(w);
			rotateBottomRight(w);
			rotateFrontCw(w);
			rotateBottomLeft(w);
			rotateFrontCcw(w);
			rotateBottomRight(w);
			rotateFrontCw(w);
		} else if (colorSide(w, 6, cornerFace(w, 2)) == LEFT_FACE) {
			rotateBottomRight(w);
			rotateBottomRight(w);
			rotateLeftDown(w);
			rotateBottomLeft(w);
			rotateLeftUp(w);
		} else if (colorSide(w, 6, cornerFace(w, 2)) == BACK_FACE) {
			rotateFrontCcw(w);
			rotateBottomRight(w);
			rotateFrontCw(w);
		}
	}

	rotateLeft(w);

	/* find the blue-red edge and place it */
	edge = findEdge(w, TOP_FACE, FRONT_FACE);
	placeEdge(w, edge, TOP_FACE);
	rotateLeft(w);

	/* find the blue-yellow edge and place it */
	edge = findEdge(w, TOP_FACE, RIGHT_FACE);
	placeEdge(w, edge, TOP_FACE);
	rotateLeft(w);

	/* find the blue-orange edge and place it */
	edge = findEdge(w, TOP_FACE, BACK_FACE);
	placeEdge(w, edge, TOP_FACE);
	rotateLeft(w);
	rotateUp(w);		/* put the blue side to the back */
}

/* This procedure swaps the the bottom front corners. */
static void
swapBottomCorners(RubikWidget w)
{
	rotateTopRight(w);
	rotateFrontCw(w);
	rotateTopLeft(w);
	rotateLeftUp(w);
	rotateTopLeft(w);
	rotateLeftDown(w);
	rotateTopRight(w);
	rotateFrontCw(w);
	rotateFrontCw(w);
}

/* This procedure places the front (BOTTOM_FACE) four corners into */
/* their correct positions. */
static void
alignCorners(RubikWidget w)
{
	int corner;

	/* find and place the green-orange-white corner (upper left) */
	corner = findCorner(w, BOTTOM_FACE, BACK_FACE, LEFT_FACE);
	if (corner == 2)
		rotateFrontCcw(w);
	else if (corner == 5)
		rotateFrontCw(w);
	else if (corner == 6) {
		rotateFrontCw(w);
		rotateFrontCw(w);
	}
	/* find and place the green-yellow-orange corner (lower left) */
	corner = findCorner(w, BOTTOM_FACE, RIGHT_FACE, BACK_FACE);
	if (corner == 2) {
		rotateCw(w);
		swapBottomCorners(w);
		rotateCcw(w);
		swapBottomCorners(w);
	} else if (corner == 6)
		swapBottomCorners(w);

	/* find and place the green-red-yellow corner (lower right) */
	corner = findCorner(w, BOTTOM_FACE, FRONT_FACE, RIGHT_FACE);
	if (corner == 2) {
		rotateCw(w);
		swapBottomCorners(w);
		rotateCcw(w);
	}
}

/* This procedure aligns the bottom corner colors (may need to be repeated). */
static void
colorAlignFront(RubikWidget w)
{
	rotateTopRight(w);
	rotateFrontCw(w);
	rotateFrontCw(w);
	rotateTopLeft(w);
	rotateFrontCw(w);
	rotateTopRight(w);
	rotateFrontCw(w);
	rotateTopLeft(w);
	rotateFrontCw(w);
	rotateFrontCw(w);
}

/* This procedure moves the remaining TOP_FACE edge piece into position */
/* from edge position #6. */
static void
alignLastEdge(RubikWidget w)
{
	/* edge piece should be in edge position #6 */
	/* check its orientation and decide which sequence to perform */
	if (colorSide(w, 1, sideFace(w, 2)) == TOP_FACE) {
		rotateCenterLeft(w);
		rotateFrontCw(w);
		rotateCenterRight(w);
		rotateFrontCcw(w);
		rotateCenterLeft(w);
		rotateFrontCcw(w);
		rotateCenterRight(w);
		rotateFrontCw(w);
	} else {
		rotateFrontCcw(w);
		rotateCenterLeft(w);
		rotateFrontCw(w);
		rotateCenterRight(w);
		rotateFrontCw(w);
		rotateCenterLeft(w);
		rotateFrontCcw(w);
	}
}

/* This procedure "shuffles" the three center edges on the front and */
/* top faces. */
static void
shuffleEdges(RubikWidget w)
{
	rotateCenterUp(w);
	rotateTopRight(w);
	rotateTopRight(w);
	rotateCenterDown(w);
	rotateTopRight(w);
	rotateTopRight(w);
}

/* This procedure should be used when the two opposing top front and back */
/* edge pieces are in position but not color aligned (this sequence is */
/* known as Rubik's Maneuver). */
static void
recolorTopEdges(RubikWidget w)
{
	rotateCenterUp(w);
	rotateTopRight(w);
	rotateCenterUp(w);
	rotateTopRight(w);
	rotateCenterUp(w);
	rotateTopRight(w);
	rotateTopRight(w);
	rotateCenterDown(w);
	rotateTopRight(w);
	rotateCenterDown(w);
	rotateTopRight(w);
	rotateCenterDown(w);
	rotateTopRight(w);
	rotateTopRight(w);
}

/* This procedure completes the BOTTOM_FACE side by color-aligning the BOTTOM_FACE */
/* corners and putting the BOTTOM_FACE edges in place. */
static void
solveBottom(RubikWidget w)
{
	int aligned;

	/* the BOTTOM_FACE corners (currently in the front) should now be in their */
	/* proper locations; next, we color align them, and then move the */
	/* bottom edges into place */
	do {
		aligned = 0;
		if (colorSide(w, 1, cornerFace(w, 0)) == BOTTOM_FACE)
			aligned++;
		if (colorSide(w, 1, cornerFace(w, 1)) == BOTTOM_FACE)
			aligned++;
		if (colorSide(w, 1, cornerFace(w, 2)) == BOTTOM_FACE)
			aligned++;
		if (colorSide(w, 1, cornerFace(w, 3)) == BOTTOM_FACE)
			aligned++;

		if (aligned == 0) {
			colorAlignFront(w);
		} else if (aligned == 1) {
			/* place aligned corner in upper right */
			if (colorSide(w, 1, cornerFace(w, 0)) == BOTTOM_FACE)
				rotateFrontCw(w);
			if (colorSide(w, 1, cornerFace(w, 2)) == BOTTOM_FACE) {
				rotateFrontCw(w);
				rotateFrontCw(w);
			}
			if (colorSide(w, 1, cornerFace(w, 3)) == BOTTOM_FACE)
				rotateFrontCcw(w);
			colorAlignFront(w);
		} else if (aligned == 2) {
			if (colorSide(w, 1, cornerFace(w, 0)) != BOTTOM_FACE)
				rotateFrontCw(w);
			else if (colorSide(w, 1, cornerFace(w, 1)) == BOTTOM_FACE)
				rotateFrontCw(w);
			if (colorSide(w, 1, cornerFace(w, 0)) != BOTTOM_FACE)
				rotateFrontCw(w);
			colorAlignFront(w);
		} else if (aligned == 3)	/* not sure if this is possible */
			colorAlignFront(w);
	} while (aligned < 4);
}

static void
solve_2_2(RubikWidget w)
{
	int i;

#if 0
	for (i = 0; i < 3; i++)	/* This is not always efficient */
		if (!checkSolved(w, False))
			rotateFrontCw(w);
#else
	if (colorSide(w, 5, cornerFace(w, 0)) == colorSide(w, 4, cornerFace(w, 1))) {
		rotateFrontCw(w);
		return;
	}
	for (i = 0; i < 2; i++) {
		if (colorSide(w, 5, cornerFace(w, 0)) == colorSide(w, 5, cornerFace(w, 2)))
			return;
		rotateFrontCcw(w);
	}
#endif
}

static void
solveBottomEdges(RubikWidget w)
{
	int edge, color;

	/* next we move the bottom edges into place */
	rotateDown(w);		/* put the green face on top */
	rotateCw(w);
	rotateCw(w);

	color = colorSide(w, 1, cornerFace(w, 0));	/* get upper front corner color */
	edge = findEdge(w, BOTTOM_FACE, color);
	placeEdge(w, edge, BOTTOM_FACE);
	rotateTopRight(w);

	color = colorSide(w, 1, cornerFace(w, 0));	/* get upper front corner color */
	edge = findEdge(w, BOTTOM_FACE, color);
	placeEdge(w, edge, BOTTOM_FACE);
	rotateTopRight(w);

	color = colorSide(w, 1, cornerFace(w, 0));	/* get upper front corner color */
	edge = findEdge(w, BOTTOM_FACE, color);
	placeEdge(w, edge, BOTTOM_FACE);
	rotateTopRight(w);

	color = colorSide(w, 1, cornerFace(w, 0));	/* get upper front corner color */
	edge = findEdge(w, BOTTOM_FACE, color);
	placeEdge(w, edge, BOTTOM_FACE);
	rotateTopRight(w);

	/* now, align the fourth blue edge piece (if necessary) */
	rotateCw(w);
	rotateCw(w);
	edge = findEdge(w, TOP_FACE, LEFT_FACE);
	if (edge == 1) {
		if (colorSide(w, 1, sideFace(w, 0)) == TOP_FACE) {
			rotateFrontCcw(w);
			rotateCenterLeft(w);
			rotateFrontCw(w);
			rotateCenterLeft(w);
			rotateFrontCw(w);
			rotateCenterLeft(w);
			rotateFrontCcw(w);
		}
	} else {
		if (edge == 5)
			rotateCenterRight(w);
		else if (edge == 7)
			rotateCenterLeft(w);
		else if (edge == 8) {
			rotateCenterRight(w);
			rotateCenterRight(w);
		}
		alignLastEdge(w);
	}
}

/* This procedure completes the solution process by placing the */
/* remaining edges in place and alignment. */
static void
alignEdges(RubikWidget w)
{
	int aligned, color, edge;

	/* move the red side to the front */
	if (colorSide(w, 1, centerFace(w)) == RIGHT_FACE)
		rotateRight(w);
	else if (colorSide(w, 1, centerFace(w)) == BACK_FACE) {
		rotateRight(w);
		rotateRight(w);
	} else if (colorSide(w, 1, centerFace(w)) == LEFT_FACE)
		rotateLeft(w);

	/* rotate the top until its aligned with the center colors */
	edge = findEdge(w, TOP_FACE, FRONT_FACE);
	if (edge == 2)
		rotateTopLeft(w);
	else if (edge == 3) {
		rotateTopLeft(w);
		rotateTopLeft(w);
	} else if (edge == 4)
		rotateTopRight(w);

	/* rotate the bottom until its aligned with the center colors */
	edge = findEdge(w, BOTTOM_FACE, FRONT_FACE);
	if (edge == 10)
		rotateBottomLeft(w);
	else if (edge == 11) {
		rotateBottomLeft(w);
		rotateBottomLeft(w);
	} else if (edge == 12)
		rotateBottomRight(w);

	if (checkSolved(w, False))
		return;

	rotateCcw(w);		/* place unaligned edges vertically */

	/* see if any edges are in correct position */
	aligned = 0;
	edge = findEdge(w, FRONT_FACE, RIGHT_FACE);
	if (edge == 1)
		aligned++;
	edge = findEdge(w, RIGHT_FACE, BACK_FACE);
	if (edge == 3)
		aligned++;
	edge = findEdge(w, BACK_FACE, LEFT_FACE);
	if (edge == 11)
		aligned++;
	edge = findEdge(w, LEFT_FACE, FRONT_FACE);
	if (edge == 9)
		aligned++;

	if (aligned == 0) {
		shuffleEdges(w);	/* put one edge into position */
		aligned++;
	}
	if (aligned == 1) {
		/* find the correct piece and move it to the back bottom edge */
		edge = findEdge(w, FRONT_FACE, RIGHT_FACE);
		if (edge == 1) {
			rotateDown(w);
			rotateDown(w);
		} else {
			edge = findEdge(w, RIGHT_FACE, BACK_FACE);
			if (edge == 3)
				rotateUp(w);
			else {
				edge = findEdge(w, LEFT_FACE, FRONT_FACE);
				if (edge == 9)
					rotateDown(w);
			}
		}

		/* shuffle */
		color = colorSide(w, 1, centerFace(w));
		if (colorSide(w, 1, sideFace(w, 3)) == color) {
			rotateRight(w);
			rotateRight(w);
			rotateDown(w);
			shuffleEdges(w);
		} else if (colorSide(w, 6, sideFace(w, 0)) == color) {
			rotateRight(w);
			rotateRight(w);
			rotateDown(w);
			shuffleEdges(w);
		} else
			shuffleEdges(w);
	}
	/* pieces should be in place; complete color alignment */
	/* find number of unaligned edge pieces and fix them */
	aligned = 0;
	if (colorSide(w, 1, sideFace(w, 0)) == colorSide(w, 1, centerFace(w)))
		aligned++;
	if (colorSide(w, 1, sideFace(w, 3)) == colorSide(w, 1, centerFace(w)))
		aligned++;
	if (colorSide(w, 3, sideFace(w, 0)) == colorSide(w, 3, centerFace(w)))
		aligned++;
	if (colorSide(w, 3, sideFace(w, 3)) == colorSide(w, 3, centerFace(w)))
		aligned++;
	if (aligned == 0) {
		recolorTopEdges(w);
		rotateDown(w);
		rotateDown(w);
		recolorTopEdges(w);
	} else if (aligned == 2) {
		if (colorSide(w, 1, sideFace(w, 0)) == colorSide(w, 1, centerFace(w)))
			do {
				rotateDown(w);
			} while (colorSide(w, 1, sideFace(w, 0)) == colorSide(w, 1, centerFace(w)));
		if (colorSide(w, 1, sideFace(w, 3)) != colorSide(w, 1, centerFace(w)))
			rotateUp(w);
		recolorTopEdges(w);
		if (!checkSolved(w, False)) {
			rotateDown(w);
			recolorTopEdges(w);
		}
	}
}

/* Wrong center rotation on top and front */
static void
shuffle2_180(RubikWidget w)
{
	int i;

	for (i = 0; i < 3; i++)
		shuffleEdges(w);
}

/* Wrong center rotation on top */
static void
shuffle1_180(RubikWidget w, Boolean Top)
{
	int i;

	if (Top) {
		for (i = 0; i < 30; i++) {
			rotateTopRight(w);
			rotateFrontCw(w);
			rotateFrontCw(w);
		}
	} else {
		for (i = 0; i < 30; i++) {
			rotateTopRight(w);
			rotateTopRight(w);
			rotateFrontCw(w);
		}
	}
}

/* Pairs off by 90 degrees */
static void
shuffle2_90(RubikWidget w, Boolean CounterTop, Boolean CounterFront)
{
	int i, num;

	num = ((CounterTop && !CounterFront) || (!CounterTop && CounterFront)) ?
		63 : 105;
	for (i = 0; i < num; i++) {
		if (CounterFront)
			rotateFrontCcw(w);
		else
			rotateFrontCw(w);
		if (CounterTop)
			rotateTopRight(w);
		else
			rotateTopLeft(w);
	}
}

static void
solveAdjCenters(RubikWidget w)
{
	int face0 = mapFace(w, 5), face1 = mapFace(w, 1);

	if ((face0 == 2 && face1 != 0) ||
			(face1 == 2 && face0 != 0)) {
		shuffle2_180(w);
	} else if (face0 == 2 && face1 == 0) {
		shuffle1_180(w, True);
	} else if (face0 == 0 && face1 == 2) {
		shuffle1_180(w, False);
	} else if (face0 == 1 && face1 == 1) {
		shuffle2_90(w, True, True);
	} else if (face0 == 3 && face1 == 1) {
		shuffle2_90(w, True, False);
	} else if (face0 == 1 && face1 == 3) {
		shuffle2_90(w, False, True);
	} else if (face0 == 3 && face1 == 3) {
		shuffle2_90(w, False, False);
	}
}

static void
solveOppCenters(RubikWidget w)
{
	int face0, face1, face2;
	Boolean fixOp = False;

	face0 = mapFace(w, 5);
	face1 = mapFace(w, 1);
	face2 = mapFace(w, 4);
	if ((face1 == 1) || (face1 == 3)) {
		rotateUp(w);
		fixOp = True;
	} else if ((face2 == 1) || (face2 == 3)) {
		rotateCw(w);
		fixOp = True;
	} else if ((face0 == 1) || (face0 == 3))
		fixOp = True;
	if (!fixOp)
		return;
	face0 = mapFace(w, 5);
	face1 = mapFace(w, 1);
	if (face0 == 1)
		shuffle2_90(w, False, True);
	else /* face0 == 3 */
		shuffle2_90(w, True, False);
	rotateUp(w);
	face0 = mapFace(w, 5);
	face1 = mapFace(w, 1);
	if (face0 == 1 && face1 == 1) {
		shuffle2_90(w, True, True);
	} else if (face0 == 3 && face1 == 1) {
		shuffle2_90(w, True, False);
	} else if (face0 == 1 && face1 == 3) {
		shuffle2_90(w, False, True);
	} else if (face0 == 3 && face1 == 3) {
		shuffle2_90(w, False, False);
	}
}

static void
solveCenters(RubikWidget w)
{
	solveAdjCenters(w); /* 5-1 */
	rotateUp(w);
	rotateUp(w);
	solveAdjCenters(w); /* 6-3 */
	rotateCw(w);
	solveAdjCenters(w); /* 4-3 */
	rotateUp(w);
	solveAdjCenters(w); /* 3-2 */
	/* only 90 degree matches left */
	rotateUp(w);
	solveAdjCenters(w); /* 2-1 */
	rotateUp(w);
	solveAdjCenters(w); /* 1-4 */
	rotateCcw(w);
	solveAdjCenters(w); /* 5-4 */
	rotateUp(w);
	solveAdjCenters(w); /* 4-6 */
	rotateUp(w);
	solveAdjCenters(w); /* 6-2 */
	rotateUp(w);
	solveAdjCenters(w); /* 2-5 */
	rotateCcw(w);
	rotateUp(w);
	solveAdjCenters(w); /* 1-6 */
	rotateUp(w);
	rotateUp(w);
	solveAdjCenters(w); /* 3-5 */
	/* only 1 90 degree matches left possible on opposite sides */
	solveOppCenters(w);
}

#ifdef REVENGE
/* Revenge */

#if 0
static void
solveFrontCorners4(RubikWidget w)
{
	/* put the orange corner on the top right */
	if (colorSide(w, 0, cornerFace(w, 2)) == BACK_FACE)
		rotateDown(w);
	else if (colorSide(w, 1, cornerFace(w, 2)) == BACK_FACE)
		rotateRight(w);
	else if (colorSide(w, 3, cornerFace(w, 2)) == BACK_FACE)
		rotateLeft(w);
	else if (colorSide(w, 4, cornerFace(w, 2)) == BACK_FACE)
		rotateUp(w);
	else if (colorSide(w, 6, cornerFace(w, 2)) == BACK_FACE) {
		rotateUp(w);
		rotateUp(w);
	}
}
#endif

static void placeBookFromWIIITop(RubikWidget w, int book, int window) {
	if (window > 0 && window < REV) {
		if (book == 0) {
			rotateTableRight(w, 0);
			rotateBookDown(w, w->rubik.sizex - 2 - window);
			rotateTableLeft(w, 0);
		} else if (book == REV) {
			rotateTableLeft(w, 0);
			rotateBookDown(w, window);
			rotateTableRight(w, 0);
		}
	} else if (window == 0) {
		rotateBookDown(w, book);
	} else if (window == REV) {
		rotateBookDown(w, book);
		rotateBookDown(w, book);
	}
}

static void placeBookFromWIIIColumn(RubikWidget w, int table, int book, int window) {
	if (table > 0 && table < REV) {
		if (book == 0) {
			if (window == REV) {
				rotateBookDown(w, 0);
			} else if (window == 0) {
				rotateBookUp(w, 0);
			}
			rotateTableRight(w, 0);
			rotateBookDown(w, table);
			rotateTableLeft(w, 0);
			if (window == REV) {
				rotateBookUp(w, 0);
			} else if (window == 0) {
				rotateBookDown(w, 0);
			}
		} else if (book == REV) {
			if (window == REV) {
				rotateBookDown(w, REV);
			} else if (window == 0) {
				rotateBookUp(w, REV);
			}
			rotateTableLeft(w, 0);
			rotateBookDown(w, w->rubik.sizex - 2 - table);
			rotateTableRight(w, 0);
			if (window == REV) {
				rotateBookUp(w, REV);
			} else if (window == 0) {
				rotateBookDown(w, REV);
			}
		}
	}
}

static void placeBookFromWIIIBottom(RubikWidget w, int book, int window) {
	if (book > 0 && book < REV &&
			window == REV) {
		rotateBookUp(w, book);
	} else if (book == 0) {
		rotateBookUp(w, REV);
		rotateBookUp(w, REV);
		rotateTableRight(w, REV);
		rotateBookUp(w, w->rubik.sizex - 2 - window);
		rotateTableLeft(w, REV);
		rotateBookDown(w, w->rubik.sizex - 2 - window);
		rotateBookDown(w, REV);
		rotateBookDown(w, REV);
	} else if (book == REV) {
		rotateBookUp(w, 0);
		rotateBookUp(w, 0);
		rotateTableLeft(w, REV);
	       rotateBookUp(w, window);
		rotateTableRight(w, REV);
		rotateBookDown(w, window);
		rotateBookDown(w, 0);
		rotateBookDown(w, 0);
	}
}

/* For size 4
 *  01         1 2
 * 2  4  =>  4     7
 * 3  5      8     11
 *  67        13 14
 */
static int sideFace4(RubikWidget w, int oct) {
	/* top left to bottom right, in groups of 2 */
	int quad = oct / 2;
	int offset = oct % 2;
	switch (quad) {
	case 0:
	case 3:
		return sideFace(w, quad) + offset;
	case 1:
	case 2:
		return (sideFace(w, quad) + offset * w->rubik.sizex);
	default:
		return w->rubik.sizex * w->rubik.sizex;
	}
}

/*static void testSideFace4(RubikWidget w) {
	for (int i = 0; i < 8; i++) {
		printf("%d %d\n", i, sideFace4(w, i));
	}
}*/


#define checkColorSide(i_face, oct, color) (colorSide(w, i_face, sideFace4(w, oct)) == color)
static RubikTBW
findEdgeTBW(RubikWidget w, int color1, int color2, int offset)
{
	RubikTBW tbw;
	tbw.table = -1, tbw.book = -1, tbw.window = -1;
	int edge = 0;
	/*edge = color1, color1 = color2, color2 = edge;*/

	for (edge = 0; edge < 12; edge++) {
		switch (edge) {
		case 0:
			if (checkColorSide(I_FRONT_FACE, offset, color1) &&
					checkColorSide(I_TOP_FACE, 6 + offset, color2)) {
			       tbw.table = 0;
				tbw.book = offset + 1;
				tbw.window = 0;
				return tbw;
			} else if (checkColorSide(I_TOP_FACE, 7 - offset, color2) &&
					checkColorSide(I_FRONT_FACE, 1 - offset, color1)) {
				tbw.table = 0;
				tbw.book = REV - offset - 1;
				tbw.window = 0;
				return tbw;
			}
			continue;
		case 1:
			if (checkColorSide(I_LEFT_FACE, 1 - offset, color1) &&
					checkColorSide(I_TOP_FACE, 3 - offset, color2)) {
				tbw.table = 0;
				tbw.book = 0;
				tbw.window = offset + 1;
				return tbw;
			} else if (checkColorSide(I_TOP_FACE, 2 + offset, color2) &&
					checkColorSide(I_LEFT_FACE, offset, color1)) {
				tbw.table = 0;
				tbw.book = 0;
				tbw.window = REV - offset - 1;
				return tbw;
			}
			continue;
		case 2:
			if (checkColorSide(I_BACK_FACE, 6 + offset, color1) &&
					checkColorSide(I_TOP_FACE, offset, color2)) {
				tbw.table = 0;
				tbw.book = offset + 1;
				tbw.window = REV;
				return tbw;
			} else if (checkColorSide(I_TOP_FACE, 1 - offset, color2) &&
					checkColorSide(I_BACK_FACE, 7 - offset, color1)) {
				tbw.table = 0;
				tbw.book = REV - offset - 1;
				tbw.window = REV;
				return tbw;
			}
			continue;
		case 3:
			if (checkColorSide(I_RIGHT_FACE, offset, color1) &&
					checkColorSide(I_TOP_FACE, 5 - offset, color2)) {
				tbw.table = 0;
				tbw.book = REV;
				tbw.window = offset + 1;
				return tbw;
			} else if (checkColorSide(I_TOP_FACE, 4 + offset, color2) &&
					checkColorSide(I_RIGHT_FACE, 1 - offset, color1)) {
				tbw.table = 0;
				tbw.book = REV;
				tbw.window = REV - offset - 1;
				return tbw;
			}
			continue;
		case 4:
			if (checkColorSide(I_FRONT_FACE, 2 + offset, color1) &&
					checkColorSide(I_LEFT_FACE, 4 + offset, color2)) {
				tbw.table = offset + 1;
				tbw.book = 0;
				tbw.window = 0;
				return tbw;
			} else if (checkColorSide(I_LEFT_FACE, 5 - offset, color1) &&
					checkColorSide(I_FRONT_FACE, 3 - offset, color2)) {
				tbw.table = REV - offset - 1;
				tbw.book = 0;
				tbw.window = 0;
				return tbw;
			}
			continue;
		case 5:
			if (checkColorSide(I_BACK_FACE, 3 - offset, color2) &&
					checkColorSide(I_LEFT_FACE, 2 + offset, color1)) {
				tbw.table = offset + 1;
				tbw.book = 0;
				tbw.window = REV;
				return tbw;
			} else if (checkColorSide(I_LEFT_FACE, 3 - offset, color2) &&
					checkColorSide(I_BACK_FACE, 2 + offset, color1)) {
				tbw.table = REV - offset - 1;
				tbw.book = 0;
				tbw.window = REV;
				return tbw;
			}
			continue;
		case 6:
			if (checkColorSide(I_BACK_FACE, 5 - offset, color1) &&
					checkColorSide(I_RIGHT_FACE, 4 + offset, color2)) {
				tbw.table = offset + 1;
				tbw.book = REV;
				tbw.window = REV;
				return tbw;
			} else if (checkColorSide(I_RIGHT_FACE, 5 - offset, color2) &&
					checkColorSide(I_BACK_FACE, 4 + offset, color1)) {
				tbw.table = REV - offset - 1;;
				tbw.book = REV;
				tbw.window = REV;
				return tbw;
			}
			continue;
		case 7:
			if (checkColorSide(I_FRONT_FACE, 4 + offset, color1) &&
					checkColorSide(I_RIGHT_FACE, 2 + offset, color2)) {
				tbw.table = offset + 1;
				tbw.book = REV;
				tbw.window = 0;
				return tbw;
			} else if (checkColorSide(I_RIGHT_FACE, 3 - offset, color2) &&
					checkColorSide(I_FRONT_FACE, 5 - offset, color1)) {
				tbw.table = REV - offset - 1;
				tbw.book = REV;
				tbw.window = 0;
				return tbw;
			}
			continue;
		case 8:
			if (checkColorSide(I_FRONT_FACE, 6 + offset, color1) &&
					checkColorSide(I_BOTTOM_FACE, offset, color2)) {
				tbw.table = REV;
				tbw.book = offset + 1;
				tbw.window = 0;
				return tbw;
			} else if (checkColorSide(I_BOTTOM_FACE, 1 - offset, color2) &&
					checkColorSide(I_FRONT_FACE, 7 - offset, color1)) {
				tbw.table = REV;
				tbw.book = REV - offset - 1;
				tbw.window = 0;
				return tbw;
			}
			continue;
		case 9:
			if (checkColorSide(I_LEFT_FACE, 7 - offset, color1) &&
					checkColorSide(I_BOTTOM_FACE, 2 + offset, color2)) {
				tbw.table = REV;
				tbw.book = 0;
				tbw.window = offset + 1;
				return tbw;
			} else if (checkColorSide(I_BOTTOM_FACE, 3 - offset, color2) &&
					checkColorSide(I_LEFT_FACE, 6 + offset, color1)) {
				tbw.table = REV;
				tbw.book = 0;
				tbw.window = REV - offset - 1;
				return tbw;
			}
			continue;
		case 10:
			if (checkColorSide(I_BACK_FACE, offset, color1) &&
					checkColorSide(I_BOTTOM_FACE, 6 + offset, color2)) {
				tbw.table = REV;
				tbw.book = offset + 1;
				tbw.window = REV;
				return tbw;
			} else if (checkColorSide(I_BOTTOM_FACE, 7 - offset, color2) &&
					checkColorSide(I_BACK_FACE, 1 - offset, color1)) {
				tbw.table = REV;
				tbw.book = REV - offset - 1;
				tbw.window = REV;
				return tbw;
			}
			continue;
		case 11:
			if (checkColorSide(I_RIGHT_FACE, 6 + offset, color1) &&
					checkColorSide(I_BOTTOM_FACE, 4 + offset, color2)) {
				tbw.table = REV;
				tbw.book = REV;
				tbw.window = offset + 1;
				return tbw;
			} else if (checkColorSide(I_BOTTOM_FACE, 5 - offset, color2) &&
					checkColorSide(I_RIGHT_FACE, 7 - offset, color1)) {
				tbw.table = REV;
				tbw.book = REV;
				tbw.window = REV - offset - 1;
				return tbw;
			}
			continue;
		}
	}
	return tbw;
}

/* This procedure finds the location of a specified edge.  An */
/* improperly-specified color combination will result in a return */
/* value of 25.  Where when first color is above second color */
/* then center is down from first color */
static int
findEdge4(RubikWidget w, int color1, int color2, int offset)
{
	int edge = 0, temp = 1/*, other = 0*/;

	do {
		/* face, cubie */
		/* Front (1) bottom edge & Bottom (6) top edge*/
		/* offset 0 or 1 */
		if (colorSide(w, 1, sideFace4(w, 6 + offset)) == color1) {
			if (colorSide(w, 6, sideFace4(w, offset)) == color2)
				edge = temp;
		} else if (colorSide(w, 1, sideFace4(w, 7 - offset)) == color2 &&
				colorSide(w, 6, sideFace4(w, 1 - offset)) == color1) {
			edge = temp;
			/*other = 1;*/
		}
		if (edge == 0) {
			if (temp < 4)
				rotateLeft(w);
			else if (temp == 4) {
				rotateLeft(w);
				rotateCw(w);
				rotateUp(w);
			} else if (temp == 8) {
				rotateUp(w);
				rotateCw(w);
			} else if (temp < 12)
			       rotateRight(w);
			else if (temp == 12) /*{
				rotateCw(w);
				rotateCw(w);
				rotateLeft(w);
			} else if (temp < 16)
				rotateRight(w);
			else if (temp == 16) {
				rotateRight(w);
				rotateCcw(w);
			} else if (temp < 20)
				rotateDown(w);
			else if (temp == 20) {
				rotateDown(w);
				rotateCcw(w);
			} else if (temp < 20)
				rotateLeft(w);
			else if (temp == 24)*/
				edge = 25;      /* end condition, just in case */
			temp++;
		}
	} while (edge == 0);

printf("Edge=%d\n", edge);
	/* put the cube back to the way it was */
	if (edge == 2) {
		rotateRight(w);
	} else if (edge == 3) {
		rotateRight(w);
		rotateRight(w);
	} else if (edge == 4) {
		rotateLeft(w);
	} else if (edge == 5) {
		rotateCcw(w);
	} else if (edge == 6) {
		rotateCcw(w);
		rotateRight(w);
	} else if (edge == 7) {
		rotateCcw(w);
		rotateRight(w);
		rotateRight(w);
	} else if (edge == 8) {
		rotateCcw(w);
		rotateLeft(w);
	} else if (edge == 9) {
		rotateCcw(w);
		rotateCcw(w);
	} else if (edge == 10) {
	}
	return edge;
}

static void
placeEdge4_0(RubikWidget w, int edge, int offset)
{
	/* first put the edge piece in position #8 (center row, left rear) */
	/* Bottom */
printf("placeEdge4_0: %d %d\n", edge, offset);
	if (edge == 1) {
		return; /* already in place */
	}
#if 0
	 else if (edge == 2) {
		rotateBookUp(w, 1 + offset);
		rotateTableRight(w, REV);
		rotateBookDown(w, 1 + offset);
		rotateTableLeft(w, REV);
	} else if (edge == 3) {
		rotateBookUp(w, REV);
		rotateBookUp(w, REV);
		rotateCw(w, 1 + offset);
		rotateCw(w, 1 + offset);
		rotateBookUp(w, REV);
		rotateBookUp(w, REV);
	} else if (edge == 4) {
		rotateBookUp(w, 1 + offset);
		rotateTableLeft(w, REV);
		rotateBookDown(w, 1 + offset);
		rotateTableRight(w, REV);
	/* Middle */
	} else if (edge == 5) {
		rotateBookDown(w, 0);
		rotateTableLeft(w, 1 + offset);
		rotateBookUp(w, 0);
	} else if (edge == 6) {
		rotateBookDown(w, 0);
		rotateTableRight(w, 1 + offset);
		rotateBookUp(w, 0);
	} else if (edge == 7) {
		rotateWindowCcw(w, 0);
		rotateBookUp(w, 0);
		rotateWindowCcw(w, w->rubik.sizex - 2 - offset);
		rotateBookDown(w, 0);
		rotateWindowCw(w, w->rubik.sizex - 2 - offset);
		rotateWindowCw(w);
	} else if (edge == 8) {
		rotateWindowCw(w, REV);
		rotateBookUp(w, 0);
		rotateWindowCw(w, w->rubik.sizex - 2 - offset);
		rotateBookDown(w, 0);
		rotateWindowCcw(w, w->rubik.sizex - 2 - offset);
		rotateWindowCcw(w, REV);
	/* Top */
	} else if (edge == 12) {
		rotateBottomRight(w);
		rotateFrontCw(w);
		rotateCenterLeft4(w, 0);
		rotateFrontCcw(w);
		rotateBottomLeft(w);
	}
#endif
}

static void
solveFirst3PairEdge3(RubikWidget w)
{
	int edge;
	edge = findEdge(w, FRONT_FACE, TOP_FACE); /* 9 */
	printf("%d FRONT_FACE TOP_FACE\n", edge);
	placeEdge(w, edge, TOP_FACE);
	/*rotateLeft(w);*/
	/*edge = findEdge(w, TOP_FACE, FRONT_FACE);
	printf("%d TOP_FACE FRONT_FACE\n", edge);
	edge = findEdge(w, BACK_FACE, TOP_FACE);
	printf("%d BACK_FACE TOP_FACE\n", edge);
	edge = findEdge(w, TOP_FACE, BACK_FACE);
	printf("%d TOP_FACE BACK_FACE\n", edge);
	edge = findEdge(w, RIGHT_FACE, TOP_FACE);
	printf("%d RIGHT_FACE TOP_FACE\n", edge);
	edge = findEdge(w, TOP_FACE, RIGHT_FACE);
	printf("%d TOP_FACE RIGHT_FACE\n", edge);*/

	/*edge = findEdge(w, FRONT_FACE, LEFT_FACE);
	printf("%d FRONT_FACE LEFT_FACE\n", edge);
	edge = findEdge(w, LEFT_FACE, FRONT_FACE);
	printf("%d LEFT_FACE FRONT_FACE\n", edge);
	edge = findEdge(w, BACK_FACE, LEFT_FACE);
	printf("%d BACK_FACE LEFT_FACE\n", edge);
	edge = findEdge(w, LEFT_FACE, BACK_FACE);
	printf("%d LEFT_FACE BACK_FACE\n", edge);
	edge = findEdge(w, RIGHT_FACE, LEFT_FACE);
	printf("%d RIGHT_FACE LEFT_FACE\n", edge);
	edge = findEdge(w, LEFT_FACE, RIGHT_FACE);
	printf("%d LEFT_FACE RIGHT_FACE\n", edge);*/
}

static void
solveFirst3PairEdge4(RubikWidget w)
{
#if 0
	printf("Front/Left\n");
	printTBW(findEdgeTBW(w, FRONT_FACE, LEFT_FACE, 0));
	printTBW(findEdgeTBW(w, FRONT_FACE, LEFT_FACE, 1));
	printf("Back/Left\n");
	printTBW(findEdgeTBW(w, BACK_FACE, LEFT_FACE, 0));
	printTBW(findEdgeTBW(w, BACK_FACE, LEFT_FACE, 1));
	printf("Back/Right\n");
	printTBW(findEdgeTBW(w, BACK_FACE, RIGHT_FACE, 0));
	printTBW(findEdgeTBW(w, BACK_FACE, RIGHT_FACE, 1));
	printf("Front/Right\n");
	printTBW(findEdgeTBW(w, FRONT_FACE, RIGHT_FACE, 0));
	printTBW(findEdgeTBW(w, FRONT_FACE, RIGHT_FACE, 1));
#endif

#if 0
	printf("Front/Top\n");
	printTBW(findEdgeTBW(w, FRONT_FACE, TOP_FACE, 0));
	printTBW(findEdgeTBW(w, FRONT_FACE, TOP_FACE, 1));
	printf("Front/Bottom\n");
	printTBW(findEdgeTBW(w, FRONT_FACE, BOTTOM_FACE, 0));
	printTBW(findEdgeTBW(w, FRONT_FACE, BOTTOM_FACE, 1));
	printf("Back/Bottom\n");
	printTBW(findEdgeTBW(w, BACK_FACE, BOTTOM_FACE, 0));
	printTBW(findEdgeTBW(w, BACK_FACE, BOTTOM_FACE, 1));
	printf("Back/Top\n");
	printTBW(findEdgeTBW(w, BACK_FACE, TOP_FACE, 0));
	printTBW(findEdgeTBW(w, BACK_FACE, TOP_FACE, 1));
#endif

#if 1
	printf("Left/Top\n");
	printTBW(findEdgeTBW(w, LEFT_FACE, TOP_FACE, 0));
	printTBW(findEdgeTBW(w, LEFT_FACE, TOP_FACE, 1));
	printf("Right/Top\n");
	printTBW(findEdgeTBW(w, RIGHT_FACE, TOP_FACE, 0));
	printTBW(findEdgeTBW(w, RIGHT_FACE, TOP_FACE, 1));
	printf("Right/Bottom\n");
	printTBW(findEdgeTBW(w, RIGHT_FACE, BOTTOM_FACE, 0));
	printTBW(findEdgeTBW(w, RIGHT_FACE, BOTTOM_FACE, 1));
	printf("Left/Bottom\n");
	printTBW(findEdgeTBW(w, LEFT_FACE, BOTTOM_FACE, 0));
	printTBW(findEdgeTBW(w, LEFT_FACE, BOTTOM_FACE, 1));
#endif
}

static void
printAllEdges(RubikWidget w)
{
	printf("Front/Top\n");
	printTBW(findEdgeTBW(w, FRONT_FACE, TOP_FACE, 0));
	printTBW(findEdgeTBW(w, FRONT_FACE, TOP_FACE, 1));
	printf("Left/Top\n");
	printTBW(findEdgeTBW(w, LEFT_FACE, TOP_FACE, 0));
	printTBW(findEdgeTBW(w, LEFT_FACE, TOP_FACE, 1));
	printf("Back/Top\n");
	printTBW(findEdgeTBW(w, BACK_FACE, TOP_FACE, 0));
	printTBW(findEdgeTBW(w, BACK_FACE, TOP_FACE, 1));
	printf("Right/Top\n");
	printTBW(findEdgeTBW(w, RIGHT_FACE, TOP_FACE, 0));
	printTBW(findEdgeTBW(w, RIGHT_FACE, TOP_FACE, 1));
	printf("Front/Left\n");
	printTBW(findEdgeTBW(w, FRONT_FACE, LEFT_FACE, 0));
	printTBW(findEdgeTBW(w, FRONT_FACE, LEFT_FACE, 1));
	printf("Back/Left\n");
	printTBW(findEdgeTBW(w, BACK_FACE, LEFT_FACE, 0));
	printTBW(findEdgeTBW(w, BACK_FACE, LEFT_FACE, 1));
	printf("Back/Right\n");
	printTBW(findEdgeTBW(w, BACK_FACE, RIGHT_FACE, 0));
	printTBW(findEdgeTBW(w, BACK_FACE, RIGHT_FACE, 1));
	printf("Front/Right\n");
	printTBW(findEdgeTBW(w, FRONT_FACE, RIGHT_FACE, 0));
	printTBW(findEdgeTBW(w, FRONT_FACE, RIGHT_FACE, 1));
	printf("Front/Bottom\n");
	printTBW(findEdgeTBW(w, FRONT_FACE, BOTTOM_FACE, 0));
	printTBW(findEdgeTBW(w, FRONT_FACE, BOTTOM_FACE, 1));
	printf("Left/Bottom\n");
	printTBW(findEdgeTBW(w, LEFT_FACE, BOTTOM_FACE, 0));
	printTBW(findEdgeTBW(w, LEFT_FACE, BOTTOM_FACE, 1));
	printf("Back/Bottom\n");
	printTBW(findEdgeTBW(w, BACK_FACE, BOTTOM_FACE, 0));
	printTBW(findEdgeTBW(w, BACK_FACE, BOTTOM_FACE, 1));
	printf("Right/Bottom\n");
	printTBW(findEdgeTBW(w, RIGHT_FACE, BOTTOM_FACE, 0));
	printTBW(findEdgeTBW(w, RIGHT_FACE, BOTTOM_FACE, 1));


#if 0
	int edge;
	edge = findEdge4(w, LEFT_FACE, FRONT_FACE , 0); /* 8 Top */
	if (edge != 8) {

	}
	printf("%d FRONT_FACE LEFT_FACE, 0\n", edge);
#endif
/* placeBookFromWIII(w, 1, 0); */
#if 0
	rotateTableRight(w, REV);
	int edge;
	edge = findEdge4(w, LEFT_FACE, BOTTOM_FACE , 0); /* 4 */
	printf("%d BOTTOM_FACE LEFT_FACE, 0\n", edge);
#endif
#if 0
	int edge;
	edge = findEdge4(w, BOTTOM_FACE, LEFT_FACE , 0); /* 16 */
	printf("%d BOTTOM_FACE LEFT_FACE, 0\n", edge);
#endif
#if 0
	int edge;
	edge = findEdge4(w, FRONT_FACE, LEFT_FACE , 0); /* 20 Bottom */
	printf("%d FRONT_FACE LEFT_FACE, 0\n", edge);
#endif
#if 0
	int edge;
	edge = findEdge4(w, FRONT_FACE, TOP_FACE , 0); /* 9 */
	printf("%d FRONT_FACE TOP_FACE, 0\n", edge);
#endif
#if 0
	int edge;
	edge = findEdge4(w, TOP_FACE, FRONT_FACE , 0); /* 21 */
	printf("%d FRONT_FACE TOP_FACE, 0\n", edge);
#endif
#if 0
	int edge;
	edge = findEdge4(w, TOP_FACE, BOTTOM_FACE , 0); /* 25 */
	printf("%d BOTTOM_FACE TOP_FACE, 0\n", edge);
#endif
/*      placeEdge4(w, edge, TOP_FACE);*/
	/*rotateLeft(w);*/

	/*edge = findEdge4(w, FRONT_FACE, TOP_FACE, 1);
	printf("%d FRONT_FACE TOP_FACE, 1\n", edge);
	placeEdge4(w, edge, TOP_FACE); */
	/*printf("%d FRONT_FACE TOP_FACE, 1\n", edge);*/
	/*placeEdge4(w, edge, FRONT_FACE);*/
	/*rotateLeft(w);*/
	/*edge = findEdge4(w, TOP_FACE, FRONT_FACE);
	printf("%d TOP_FACE FRONT_FACE\n", edge);
	edge = findEdge4(w, BACK_FACE, TOP_FACE);
	printf("%d BACK_FACE TOP_FACE\n", edge);
	edge = findEdge4(w, TOP_FACE, BACK_FACE);
	printf("%d TOP_FACE BACK_FACE\n", edge);
	edge = findEdge4(w, RIGHT_FACE, TOP_FACE);
	printf("%d RIGHT_FACE TOP_FACE\n", edge);
	edge = findEdge4(w, TOP_FACE, RIGHT_FACE);
	printf("%d TOP_FACE RIGHT_FACE\n", edge);*/


/*
	edge = findEdge4(w, FRONT_FACE, LEFT_FACE);
	printf("%d FRONT_FACE LEFT_FACE\n", edge);
	edge = findEdge4(w, LEFT_FACE, FRONT_FACE);
	printf("%d LEFT_FACE FRONT_FACE\n", edge);
	edge = findEdge4(w, BACK_FACE, LEFT_FACE);
	printf("%d BACK_FACE LEFT_FACE\n", edge);
	edge = findEdge4(w, LEFT_FACE, BACK_FACE);
	printf("%d LEFT_FACE BACK_FACE\n", edge);
	edge = findEdge4(w, RIGHT_FACE, LEFT_FACE);
	printf("%d RIGHT_FACE LEFT_FACE\n", edge);
	edge = findEdge4(w, LEFT_FACE, RIGHT_FACE);
	printf("%d LEFT_FACE RIGHT_FACE\n", edge);*/
}

static void
solve6PairEdge4(RubikWidget w)
{
	/* IIIA TOP_FACE-FRONT_FACE, TOP_FACE-BACK_FACE TOP_FACE-RIGHT_FACE */
	solveFirst3PairEdge4(w);
	/* IIIB LEFT_FACE-FRONT_FACE, LEFT_FACE-BACK_FACE LEFT_FACE-RIGHT_FACE */
	/*solveSecond3PairEdge4(w); */
}
#endif

void
printSomePieces(RubikWidget w)
{
#ifdef REVENGE
/* using this for debugging */
	if (w->rubik.sizex == 3) {
		(void) printf("printSomePieces: 3\n");
		solveFirst3PairEdge3(w);
	}
	if (w->rubik.sizex == 4) {
		(void) printf("printSomePieces: 4\n");
		solveFirst3PairEdge4(w);
	}
	/*rotateTopLeft(w);*/
#endif
}

/* This procedure coordinates the solution process. */
void
solveSomePieces(RubikWidget w)
{
	setPuzzle(w, ACTION_RESET);
	if (solvingFlag)
		return;
#ifdef JMP
	if (!setjmp(solve_env))
#endif
	{
		solvingFlag = True;

		if (!checkSolved(w, w->rubik.orient)) {
			solveTop(w);
			alignCorners(w);
			solveBottom(w);
			if (w->rubik.sizex % 2 == 1) {
				solveBottomEdges(w);
				alignEdges(w);
				if (w->rubik.orient)
					solveCenters(w);
			} else {
				solve_2_2(w);
#ifdef REVENGE
				if (w->rubik.sizex > 2) {
					/* 4x4x4 */
					solve6PairEdge4(w);
					/*solve2PairEdge4(w);
					solve4PairEdge4(w);
					solveCenter4(w);*/
				}
#endif
			}
		}
	}
#ifdef JMP
	else {
		drawAllPieces(w);
	}
	abortSolvingFlag = False;
#endif
	solvingFlag = False;
	w->rubik.cheat = True; /* Assume the worst. */
	setPuzzle(w, ACTION_CHEAT);
	setPuzzle(w, ACTION_COMPUTED);
}
